<div id="indexlinkbox">
    <h3 id="indexlinklabel">{{ _('Index') }}</h3>

    <div class="indexlinkwrapper nav-menu">
        <ul>
            <li class="toctree-l1">
                <a href="{{ pathto('genindex')|e }}" class="indexlink genindex">
                    {{ _('Alphabetical') }}
                </a>
            </li>
            <li class="toctree-l1">
                <a href="{{ pathto('gentree')|e }}" class="indexlink gentree">
                    {{ _('Tree') }}
                </a>
            </li>
        </ul>
    </div>
</div>
