<div class="nav-container">
    <aside class="nav">
      <div class="panels">
        <div class="nav-panel-menu is-active" data-panel="menu">
            <nav class="nav-menu">
                {{ toctree(includehidden=theme_sidebar_includehidden, collapse=theme_sidebar_collapse) }}
            </nav>
        </div>
      </div>
    </aside>
</div>
